@@ -25,20 +25,20 @@
  * for Mupen64Plus v2.0. 
  */
  
-#include <stdio.h>
-#include <string.h>
-#include <stdlib.h>
 #include <stdarg.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 
-#include <SDL_main.h>
-
+#include "SDL_main.h"
 #include "cheat.h"
+#include "compare_core.h"
+#include "core_interface.h"
+#include "m64p_types.h"
 #include "main.h"
+#include "osal_preproc.h"
 #include "plugin.h"
 #include "version.h"
-#include "core_interface.h"
-#include "compare_core.h"
-#include "osal_preproc.h"
 
 /* Version number for UI-Console config section parameters */
 #define CONFIG_PARAM_VERSION     1.00
@@ -54,6 +54,7 @@ static m64p_handle l_ConfigUI = NULL;
 static const char *l_CoreLibPath = NULL;
 static const char *l_ConfigDirPath = NULL;
 static const char *l_ROMFilepath = NULL;       // filepath of ROM to load & run at startup
+static const char *l_SaveStatePath = NULL;     // save state to load at startup
 
 #if defined(SHAREDIR)
   static const char *l_DataDirPath = SHAREDIR;
@@ -88,6 +89,23 @@ void DebugMessage(int level, const char *message, ...)
 
 void DebugCallback(void *Context, int level, const char *message)
 {
+#ifdef ANDROID
+    if (level == M64MSG_ERROR)
+        __android_log_print(ANDROID_LOG_ERROR, (const char *) Context, "%s", message);
+    else if (level == M64MSG_WARNING)
+        __android_log_print(ANDROID_LOG_WARN, (const char *) Context, "%s", message);
+    else if (level == M64MSG_INFO)
+        __android_log_print(ANDROID_LOG_INFO, (const char *) Context, "%s", message);
+    else if (level == M64MSG_STATUS)
+        __android_log_print(ANDROID_LOG_DEBUG, (const char *) Context, "%s", message);
+    else if (level == M64MSG_VERBOSE)
+    {
+        if (g_Verbose)
+            __android_log_print(ANDROID_LOG_VERBOSE, (const char *) Context, "%s", message);
+    }
+    else
+        __android_log_print(ANDROID_LOG_ERROR, (const char *) Context, "Unknown: %s", message);
+#else
     if (level == M64MSG_ERROR)
         printf("%s Error: %s\n", (const char *) Context, message);
     else if (level == M64MSG_WARNING)
@@ -103,10 +121,20 @@ void DebugCallback(void *Context, int level, const char *message)
     }
     else
         printf("%s Unknown: %s\n", (const char *) Context, message);
+#endif
 }
 
 static void FrameCallback(unsigned int FrameIndex)
 {
+    // load savestate at first frame if needed
+    if (l_SaveStatePath != NULL && FrameIndex == 0)
+    {
+        if ((*CoreDoCommand)(M64CMD_STATE_LOAD, 0, (void *) l_SaveStatePath) != M64ERR_SUCCESS)
+        {
+            DebugMessage(M64MSG_WARNING, "couldn't load state, rom will run normally.");
+        }
+    }
+
     // take a screenshot if we need to
     if (l_TestShotList != NULL)
     {
@@ -125,6 +153,7 @@ static void FrameCallback(unsigned int FrameIndex)
         }
     }
 }
+
 /*********************************************************************************************************
  *  Configuration handling
  */
@@ -224,39 +253,40 @@ static void printUsage(const char *progname)
     printf("Usage: %s [parameters] [romfile]\n"
            "\n"
            "Parameters:\n"
-           "    --noosd               : disable onscreen display\n"
-           "    --osd                 : enable onscreen display\n"
-           "    --fullscreen          : use fullscreen display mode\n"
-           "    --windowed            : use windowed display mode\n"
-           "    --resolution (res)    : display resolution (640x480, 800x600, 1024x768, etc)\n"
-           "    --nospeedlimit        : disable core speed limiter (should be used with dummy audio plugin)\n"
-           "    --cheats (cheat-spec) : enable or list cheat codes for the given rom file\n"
-           "    --corelib (filepath)  : use core library (filepath) (can be only filename or full path)\n"
-           "    --configdir (dir)     : force configation directory to (dir); should contain mupen64plus.cfg\n"
-           "    --datadir (dir)       : search for shared data files (.ini files, languages, etc) in (dir)\n"
-           "    --plugindir (dir)     : search for plugins in (dir)\n"
-           "    --sshotdir (dir)      : set screenshot directory to (dir)\n"
-           "    --gfx (plugin-spec)   : use gfx plugin given by (plugin-spec)\n"
-           "    --audio (plugin-spec) : use audio plugin given by (plugin-spec)\n"
-           "    --input (plugin-spec) : use input plugin given by (plugin-spec)\n"
-           "    --rsp (plugin-spec)   : use rsp plugin given by (plugin-spec)\n"
-           "    --emumode (mode)      : set emu mode to: 0=Pure Interpreter 1=Interpreter 2=DynaRec\n"
-           "    --testshots (list)    : take screenshots at frames given in comma-separated (list), then quit\n"
-           "    --set (param-spec)    : set a configuration variable, format: ParamSection[ParamName]=Value\n"
-           "    --core-compare-send   : use the Core Comparison debugging feature, in data sending mode\n"
-           "    --core-compare-recv   : use the Core Comparison debugging feature, in data receiving mode\n"
-           "    --nosaveoptions       : do not save the given command-line options in configuration file\n"
-           "    --verbose             : print lots of information\n"
-           "    --help                : see this help message\n\n"
+           "    --noosd                : disable onscreen display\n"
+           "    --osd                  : enable onscreen display\n"
+           "    --fullscreen           : use fullscreen display mode\n"
+           "    --windowed             : use windowed display mode\n"
+           "    --resolution (res)     : display resolution (640x480, 800x600, 1024x768, etc)\n"
+           "    --nospeedlimit         : disable core speed limiter (should be used with dummy audio plugin)\n"
+           "    --cheats (cheat-spec)  : enable or list cheat codes for the given rom file\n"
+           "    --corelib (filepath)   : use core library (filepath) (can be only filename or full path)\n"
+           "    --configdir (dir)      : force configation directory to (dir); should contain mupen64plus.cfg\n"
+           "    --datadir (dir)        : search for shared data files (.ini files, languages, etc) in (dir)\n"
+           "    --plugindir (dir)      : search for plugins in (dir)\n"
+           "    --sshotdir (dir)       : set screenshot directory to (dir)\n"
+           "    --gfx (plugin-spec)    : use gfx plugin given by (plugin-spec)\n"
+           "    --audio (plugin-spec)  : use audio plugin given by (plugin-spec)\n"
+           "    --input (plugin-spec)  : use input plugin given by (plugin-spec)\n"
+           "    --rsp (plugin-spec)    : use rsp plugin given by (plugin-spec)\n"
+           "    --emumode (mode)       : set emu mode to: 0=Pure Interpreter 1=Interpreter 2=DynaRec\n"
+           "    --savestate (filepath) : savestate loaded at startup\n"
+           "    --testshots (list)     : take screenshots at frames given in comma-separated (list), then quit\n"
+           "    --set (param-spec)     : set a configuration variable, format: ParamSection[ParamName]=Value\n"
+           "    --core-compare-send    : use the Core Comparison debugging feature, in data sending mode\n"
+           "    --core-compare-recv    : use the Core Comparison debugging feature, in data receiving mode\n"
+           "    --nosaveoptions        : do not save the given command-line options in configuration file\n"
+           "    --verbose              : print lots of information\n"
+           "    --help                 : see this help message\n\n"
            "(plugin-spec):\n"
-           "    (pluginname)          : filename (without path) of plugin to find in plugin directory\n"
-           "    (pluginpath)          : full path and filename of plugin\n"
-           "    'dummy'               : use dummy plugin\n\n"
+           "    (pluginname)           : filename (without path) of plugin to find in plugin directory\n"
+           "    (pluginpath)           : full path and filename of plugin\n"
+           "    'dummy'                : use dummy plugin\n\n"
            "(cheat-spec):\n"
-           "    'list'                : show all of the available cheat codes\n"
-           "    'all'                 : enable all of the available cheat codes\n"
-           "    (codelist)            : a comma-separated list of cheat code numbers to enable,\n"
-           "                            with dashes to use code variables (ex 1-2 to use cheat 1 option 2)\n"
+           "    'list'                 : show all of the available cheat codes\n"
+           "    'all'                  : enable all of the available cheat codes\n"
+           "    (codelist)             : a comma-separated list of cheat code numbers to enable,\n"
+           "                             with dashes to use code variables (ex 1-2 to use cheat 1 option 2)\n"
            "\n", progname);
 
     return;
@@ -529,6 +559,11 @@ static m64p_error ParseCommandLineFinal(int argc, const char **argv)
             }
             (*ConfigSetParameter)(l_ConfigCore, "R4300Emulator", M64TYPE_INT, &emumode);
         }
+        else if (strcmp(argv[i], "--savestate") == 0 && ArgsLeft >= 1)
+        {
+            l_SaveStatePath = argv[i+1];
+            i++;
+        }
         else if (strcmp(argv[i], "--testshots") == 0 && ArgsLeft >= 1)
         {
             l_TestShotList = ParseNumberList(argv[i+1], NULL);
@@ -577,6 +612,24 @@ static m64p_error ParseCommandLineFinal(int argc, const char **argv)
 /*********************************************************************************************************
 * main function
 */
+
+/* Allow state callback in external module to be specified via build flags (header and function name) */
+#ifdef CALLBACK_HEADER
+#define xstr(s) str(s)
+#define str(s) #s
+#include xstr(CALLBACK_HEADER)
+#endif
+
+#ifndef CALLBACK_FUNC
+#define CALLBACK_FUNC NULL
+#endif
+
+#ifndef WIN32
+/* Allow external modules to call the main function as a library method.  This is useful for user
+ * interfaces that simply layer on top of (rather than re-implement) UI-Console (e.g. mupen64plus-ae).
+ */
+__attribute__ ((visibility("default")))
+#endif
 int main(int argc, char *argv[])
 {
     int i;
@@ -598,7 +651,7 @@ int main(int argc, char *argv[])
         return 2;
 
     /* start the Mupen64Plus core library, load the configuration file */
-    m64p_error rval = (*CoreStartup)(CORE_API_VERSION, l_ConfigDirPath, l_DataDirPath, "Core", DebugCallback, NULL, NULL);
+    m64p_error rval = (*CoreStartup)(CORE_API_VERSION, l_ConfigDirPath, l_DataDirPath, "Core", DebugCallback, NULL, CALLBACK_FUNC);
     if (rval != M64ERR_SUCCESS)
     {
         DebugMessage(M64MSG_ERROR, "couldn't start Mupen64Plus core library.");
@@ -716,12 +769,12 @@ int main(int argc, char *argv[])
         }
     }
 
-    /* set up Frame Callback if --testshots is enabled */
-    if (l_TestShotList != NULL)
+    /* set up Frame Callback if --testshots or --savestate is enabled */
+    if (l_TestShotList != NULL || l_SaveStatePath != NULL)
     {
         if ((*CoreDoCommand)(M64CMD_SET_FRAME_CALLBACK, 0, FrameCallback) != M64ERR_SUCCESS)
         {
-            DebugMessage(M64MSG_WARNING, "couldn't set frame callback, so --testshots won't work.");
+            DebugMessage(M64MSG_WARNING, "couldn't set frame callback, --testshots and/or --savestate will not work.");
         }
     }
 
